import shutil
from pathlib import Path
from typing import List, Optional, Union

from packaging.requirements import Requirement
from packaging.specifiers import SpecifierSet
from packaging.version import Version, parse
from pydantic import BaseModel, ConfigDict, field_validator


class PackageSpec(BaseModel):
    package_name: str
    root_name_override: Optional[str] = None
    version: Version = Version("0.0.1")
    requirements: List[Requirement] = []

    @field_validator("version", mode="before")
    @classmethod
    def _validate_version(cls, v: Union[str, Version]):
        if isinstance(v, Version):
            return v
        return parse(v)

    @property
    def root_name(self) -> str:

        if self.root_name_override is None:
            return self.package_name
        return self.root_name_override

    @classmethod
    def _parse_requirement(
        cls, package: str, specifier: Union[Version, SpecifierSet, None] = None
    ) -> Requirement:
        if specifier is None:
            specifier = SpecifierSet()
        elif isinstance(specifier, Version):
            specifier = SpecifierSet(f"=={specifier}")
        return Requirement(f"{package}{specifier}")

    def add_requirement(
        self, package: str, specifier: Union[Version, SpecifierSet, None] = None
    ) -> Requirement:
        req = self._parse_requirement(package, specifier)
        self.requirements.append(req)
        return req

    model_config = ConfigDict(arbitrary_types_allowed=True)


class PackageBuilder:
    def __init__(
        self,
        parent_dir: Path,
        spec: PackageSpec,
    ):
        self.parent_dir = (
            parent_dir if isinstance(parent_dir, Path) else Path(parent_dir)
        )
        self.spec = spec.model_copy(deep=True)

    @property
    def root_path(self) -> Path:
        return self.parent_dir / self.spec.root_name

    @property
    def package_path(self) -> Path:
        return self.root_path / self.spec.package_name

    def write_setup(self):
        setup_path = self.root_path / "setup.py"
        setup_path.write_text(f"""#!/usr/bin/env python
from setuptools import setup
from pathlib import Path

PWD = Path(__file__).parent
root = PWD.resolve()

about_path = root / {repr(self.spec.package_name)} / "about.py"
with about_path.open("r", encoding="utf8") as f:
    about = {{}}
    exec(f.read(), about)

requirements_path = root / "requirements.in"
with requirements_path.open("r", encoding="utf8") as f:
    requirements = [line.strip() for line in f]

setup(name=about["__name__"],
      version=about["__version__"],
      description=about["__summary__"],
      url=about["__url__"],
      author=about["__author__],
      author_email=about["__email__"],
      license='MIT',
      packages=[about["__name__"]],
      zip_safe=False)
""")

    def write_about(self):
        about_path = self.package_path / "about.py"
        about_path.write_text(f"""__title__ = {repr(self.spec.package_name)}
__name__ = {repr(self.spec.package_name)}
__version__ = {repr(str(self.spec.version))}
__summary__ = 'Mock package'
__email__ = 'contact@explosion.ai'
__author__ = 'Explosion'
__uri__ = 'https://explosion.rocks'
__license__ = 'MIT'
            """)

    def write_requirements(self):
        requirements_path = self.root_path / "requirements.in"
        requirements_path.write_text(
            "\n".join([str(r) for r in self.spec.requirements])
        )

    def write_init(self):
        init_path = self.package_path / "__init__.py"
        init_path.write_text("""from .about import *""")

    def write(self, exists_ok: bool = False):
        self.root_path.mkdir(exist_ok=exists_ok)
        self.package_path.mkdir(exist_ok=exists_ok)
        self.write_setup()
        self.write_about()
        self.write_requirements()
        self.write_init()

    def delete(self):
        shutil.rmtree(self.root_path)

    def __enter__(self):
        self.write()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.delete()



"""Shared infrastructure helpers for the infra command group."""

import base64
import json
import subprocess
import tempfile
from pathlib import Path
from uuid import UUID

from wasabi import msg

from prodigy_teams_pam_sdk import Client
from prodigy_teams_pam_sdk.models import BrokerCredentials


def get_cluster_credentials(project: str, zone: str, cluster_name: str) -> None:
    """Run gcloud container clusters get-credentials to configure kubectl."""
    subprocess.run(
        [
            "gcloud",
            "container",
            "clusters",
            "get-credentials",
            cluster_name,
            "--zone",
            zone,
            "--project",
            project,
        ],
        check=True,
        text=True,
    )


def ensure_broker_keypair(
    infra_secret_name: str, namespace: str = "prodigy-teams"
) -> str:
    """Ensure the infra K8s secret has a broker RSA keypair, return the public PEM.

    If the secret already exists and contains a private key, the existing public
    key is returned.  Otherwise a new keypair is generated and the secret is
    created/patched.
    """
    from ...key_pair import generate_key_pair

    check = subprocess.run(
        [
            "kubectl",
            "get",
            "secret",
            infra_secret_name,
            "-n",
            namespace,
            "-o",
            "jsonpath={.data.PRODIGY_TEAMS_PUBLIC_KEY}",
        ],
        capture_output=True,
        text=True,
    )
    if check.returncode == 0 and check.stdout.strip():
        public_pem = base64.b64decode(base64.b64decode(check.stdout.strip())).decode(
            "utf-8"
        )
        msg.info(f"Using existing broker keypair from secret {infra_secret_name}")
        return public_pem

    msg.info("Generating new broker RSA keypair...")
    kp = generate_key_pair()
    private_key_b64 = base64.b64encode(kp.private_pem.encode()).decode()
    public_key_b64 = base64.b64encode(kp.public_pem.encode()).decode()
    cmd = [
        "kubectl",
        "create",
        "secret",
        "generic",
        infra_secret_name,
        f"--namespace={namespace}",
        f"--from-literal=PRODIGY_TEAMS_PRIVATE_KEY={private_key_b64}",
        f"--from-literal=PRODIGY_TEAMS_PUBLIC_KEY={public_key_b64}",
        "--dry-run=client",
        "-o",
        "yaml",
    ]
    create = subprocess.run(cmd, capture_output=True, text=True, check=True)
    subprocess.run(
        ["kubectl", "apply", "-f", "-"],
        input=create.stdout,
        check=True,
        text=True,
    )
    msg.good(f"Created broker keypair in secret {infra_secret_name}")
    return kp.public_pem


def k3s_traefik_exists() -> bool:
    """Check if K3s's bundled Traefik is already running in kube-system."""
    result = subprocess.run(
        ["kubectl", "get", "deployment", "traefik", "-n", "kube-system"],
        capture_output=True,
        text=True,
    )
    return result.returncode == 0


def ensure_traefik(public_ip: str = "") -> None:
    """Install or upgrade Traefik via Helm."""
    if k3s_traefik_exists():
        msg.info(
            "K3s bundled Traefik detected in kube-system — skipping install. "
            "Customise via HelmChartConfig in /var/lib/rancher/k3s/server/manifests/."
        )
        return
    msg.info("Installing / upgrading Traefik...")
    subprocess.run(
        ["helm", "repo", "add", "traefik", "https://traefik.github.io/charts"],
        capture_output=True,
        text=True,
    )
    subprocess.run(
        ["helm", "repo", "update", "traefik"],
        capture_output=True,
        text=True,
    )
    values: dict = {
        "providers": {
            "kubernetesIngress": {"enabled": True},
        },
        "ingressRoute": {"dashboard": {"enabled": False}},
        "ports": {"web": {"port": 80}, "websecure": {"port": 443}},
        "securityContext": {
            "capabilities": {"drop": [], "add": ["NET_BIND_SERVICE"]},
        },
    }
    if public_ip:
        values["service"] = {"spec": {"loadBalancerIP": public_ip}}
    with tempfile.NamedTemporaryFile(mode="w", suffix=".json", delete=False) as tmp:
        tmp.write(json.dumps(values))
        tmp_path = tmp.name
    try:
        subprocess.run(
            [
                "helm",
                "upgrade",
                "--install",
                "traefik",
                "traefik/traefik",
                "-n",
                "traefik",
                "--create-namespace",
                "--skip-crds",
                "-f",
                tmp_path,
            ],
            check=True,
            text=True,
        )
    finally:
        Path(tmp_path).unlink(missing_ok=True)
    if public_ip:
        msg.good(f"Traefik installed with static IP {public_ip}")
    else:
        msg.info(
            "Traefik installed. Get the LoadBalancer IP with:\n"
            "  kubectl get svc -n traefik traefik"
            " -o jsonpath='{.status.loadBalancer.ingress[0].ip}'"
        )


def ensure_cert_manager() -> None:
    """Install cert-manager via Helm if not already present."""
    check = subprocess.run(
        ["helm", "status", "cert-manager", "-n", "cert-manager"],
        capture_output=True,
        text=True,
    )
    if check.returncode == 0:
        return
    msg.info("Installing cert-manager...")
    subprocess.run(
        ["helm", "repo", "add", "jetstack", "https://charts.jetstack.io"],
        check=True,
        text=True,
    )
    subprocess.run(["helm", "repo", "update", "jetstack"], check=True, text=True)
    subprocess.run(
        [
            "helm",
            "upgrade",
            "--install",
            "cert-manager",
            "jetstack/cert-manager",
            "-n",
            "cert-manager",
            "--create-namespace",
            "--set",
            "crds.enabled=true",
            "--set",
            "crds.keep=true",
        ],
        check=True,
        text=True,
    )


def get_cluster_auth(client: Client, domain: str) -> BrokerCredentials:
    """Fetch broker credentials from PAM for a given cluster domain."""
    broker = client.broker.read(address=domain)
    return client.broker.credentials(broker_id=broker.id)


def ensure_cluster_record(
    client: Client,
    org_id: UUID,
    domain: str,
    public_key: str,
    *,
    cluster_name: str = "Test Cluster",
    cloud_provider: str = "gcp",
) -> None:
    """Create or update the cluster record in PAM."""
    import pydantic
    from pydantic import AnyUrl

    from prodigy_teams_pam_sdk.errors import BrokerNotFound
    from prodigy_teams_pam_sdk.models import BrokerCreating, BrokerUpdating

    from ...util import URL

    try:
        broker = client.broker.read(address=domain)
    except BrokerNotFound:
        body = BrokerCreating(
            name=cluster_name,
            org_id=org_id,
            address=pydantic.TypeAdapter(AnyUrl).validate_python(URL.parse(domain).url),
            cloud_provider=cloud_provider,
            public_key=public_key,
            state="creating",
            cloud={},
        )
        _ = client.broker.create(body)
    else:
        body = BrokerUpdating(
            id=broker.id,
            name=cluster_name,
            address=domain,
            cloud_provider=cloud_provider,
            public_key=public_key,
            state="creating",
            cloud={},
        )
        _ = client.broker.update(body)



import functools
import posixpath
from pathlib import Path, PurePosixPath
from typing import Optional, Protocol

from radicli import Arg
from wasabi import msg

from prodigy_teams_broker_sdk import Client as BrokerClient
from prodigy_teams_broker_sdk import models as broker_models

from ...cli import cli
from ...errors import BrokerError, CLIError
from ...messages import Messages
from ...ui import Progress, is_silent, print_mutation_result
from ...util import _is_root_path, _resolve_broker_ref, is_local, resolve_remote_path
from .._state import get_auth_state, get_saved_settings


class CopyCallable(Protocol):
    def __call__(
        self,
        *,
        src: str,
        dest: str,
        overwrite: bool,
        recurse: bool,
        make_dirs: bool,
        broker_client: BrokerClient,
    ) -> str | None: ...


def _infer_remote_source_base(src: str, listed_paths: list[str]) -> str:
    """Recover the absolute remote source root from a builtin-alias request."""
    if posixpath.isabs(src) or not listed_paths:
        return src

    common_path = posixpath.commonpath(listed_paths)
    suffix = src.rstrip("/")
    if suffix.startswith("{"):
        _, _, suffix = suffix.partition("}")
        suffix = suffix.lstrip("/")
    if not suffix:
        return common_path

    suffix_parts = PurePosixPath(suffix).parts
    common = PurePosixPath(common_path)
    for candidate in (common, *common.parents):
        if len(candidate.parts) < len(suffix_parts):
            continue
        if candidate.parts[-len(suffix_parts) :] == suffix_parts:
            return str(candidate)
    return common_path


@cli.subcommand(
    "files",
    "cp",
    src=Arg(help=Messages.remote_local_path.format(noun="source")),
    dest=Arg(help=Messages.remote_local_path.format(noun="destination")),
    recurse=Arg("--recurse", "-r", help=Messages.recurse_copy),
    make_dirs=Arg("--make-dirs", help=Messages.make_dirs),
    overwrite=Arg("--overwrite", help=Messages.overwrite),
    cluster_host=Arg("--cluster-host", help=Messages.cluster_host),
    as_json=Arg("--json", help=Messages.as_json),
)
def cp(
    src: str,
    dest: str,
    recurse: bool = False,
    make_dirs: bool = False,
    overwrite: bool = False,
    cluster_host: Optional[str] = None,
    as_json: bool = False,
) -> None:
    """Copy files to and from the cluster"""
    settings = get_saved_settings()
    auth = get_auth_state()
    broker_host = str(
        _resolve_broker_ref(auth.client, cluster_host or settings.broker_host)
    )

    is_src_local = is_local(src)
    is_dest_local = is_local(dest)

    def _cp(f: CopyCallable, *, src: str, dest: str) -> str | None:
        return f(
            src=src,
            dest=dest,
            overwrite=overwrite,
            recurse=recurse,
            make_dirs=make_dirs,
            broker_client=auth.broker_client,
        )

    if is_src_local and is_dest_local:  # both local
        raise CLIError(Messages.E015.format(verb="copy"), f"{src}, {dest}")
    elif is_src_local and not is_dest_local:  # from local to remote
        err = _cp(
            _upload,
            src=src,
            dest=resolve_remote_path(auth.client, dest, broker_host),
        )
    elif not is_src_local and is_dest_local:
        # From remote to local
        err = _cp(
            _download,
            src=resolve_remote_path(auth.client, src, broker_host),
            dest=dest,
        )
    else:
        # Both remote
        err = _cp(
            _copy_on_remote,
            src=resolve_remote_path(auth.client, src, broker_host),
            dest=resolve_remote_path(auth.client, dest, broker_host),
        )
    if err is not None:
        raise CLIError(err)
    elif as_json:
        print_mutation_result(
            {"status": "ok", "src": src, "dest": dest},
            Messages.T011,
            as_json=True,
        )
    elif not is_silent():
        msg.good(Messages.T011, f"{src} -> {dest}")


def _upload(
    *,
    src: str,
    dest: str,
    make_dirs: bool,
    overwrite: bool,
    recurse: bool,
    broker_client: BrokerClient,
) -> str | None:
    src_has_trailing_slash = src.endswith("/")
    src_path = Path(src)
    src_base = src

    @functools.lru_cache()
    def _ls(remote_path: str):
        try:
            files = broker_client.files.list_dir(
                broker_models.Listing(
                    path=remote_path, recurse=False, include_stats=False
                )
            )
            return files
        except BrokerError as e:
            raise CLIError(Messages.E018, e)

    src_exists = src_path.exists()
    src_is_file = src_path.is_file()
    src_is_dir = src_path.is_dir()

    dest_exists = _ls(dest).exists
    dest_is_file = _ls(dest).is_file
    dest_is_dir = dest_exists and not dest_is_file

    # print(
    #     f"""
    # _upload({src=},{dest=}):
    #            {_ls(dest).paths=}
    #            {src_is_file=}
    #            {src_is_dir=}
    #            {src_has_trailing_slash=}
    #
    #            {dest_exists=}
    #            {dest_is_dir=}
    #            {dest_is_file=}
    # """.strip(),
    #     file=sys.stderr,
    # )

    if not src_exists:
        raise CLIError("No such file or directory", src)

    if src_is_dir and not recurse:
        raise CLIError(Messages.E016.format(verb="upload"), src_path)

    if dest_is_file:
        if src_is_file and not overwrite:
            raise CLIError("The destination file already exists and overwrite is False")
        elif src_is_dir:
            raise CLIError("A file already exists at the destination path")

    if src_is_file:
        src_file = src_path
        if dest_is_file:
            dest_file = dest
        elif dest_is_dir or dest.endswith("/"):
            dest_file = posixpath.join(dest, src_path.name)
        else:
            dest_file = dest
        dest_base = posixpath.dirname(dest_file)
        create_base = False
        paths = [(src_file, dest_file)]
    else:
        if dest_is_dir and src_has_trailing_slash:
            # `cp src/ dest_exists/` copies `src/**` into `dest_exists/**`
            dest_base = dest
            create_base = False
        elif dest_is_dir and not src_has_trailing_slash:
            # `cp src dest_exists/` copies `src/**` into `dest_exists/src/**`
            dest_base = posixpath.join(dest, src_path.name)
            create_base = True
        else:
            # `cp src dest_noexists/` copies `src/**` into `dest_noexists/**`
            create_base = True
            dest_base = dest
        src_base = src_path

        paths = [
            (
                Path(src_file),
                posixpath.join(
                    dest_base, str(PurePosixPath(src_file.relative_to(src_base)))
                ),
            )
            for src_file in src_path.glob("**/*")
        ]

    if not make_dirs and not _is_root_path(dest_base):
        # we can't precreate empty dirs on cloud paths, but we can
        # override the make_dirs flag to create the appropriate dirs on demand
        # if we first check the existence of the parent dirs
        required_parent = posixpath.dirname(dest_base) if create_base else dest_base

        parent_ls = _ls(required_parent)

        # print(
        #     f"""
        #        {required_parent=}
        #        {parent_ls.exists=}
        #        {parent_ls.is_file=}
        #        {create_base=}
        # """.rstrip(),
        #     file=sys.stderr,
        # )
        if parent_ls.exists and not parent_ls.is_file:
            make_dirs = True
        elif parent_ls.is_file:
            raise CLIError(
                "A file already exists at the destination path", required_parent
            )
        else:
            raise CLIError(Messages.E021, dest_base)

    #     print(
    #         f"""
    #                {paths=}
    # """.rstrip(),
    #         file=sys.stderr,
    #     )
    with Progress(paths) as paths:
        for src_file, dest_file in paths:
            if src_file.is_dir():
                continue
            body = src_file.open("rb")
            try:
                broker_client.files.upload(
                    body, dest=dest_file, overwrite=overwrite, make_dirs=make_dirs
                )
            except BrokerError as e:
                raise CLIError(Messages.E017.format(verb="upload"), e)


def _download(
    *,
    src: str,
    dest: str | Path,
    make_dirs: bool,
    overwrite: bool,
    recurse: bool,
    broker_client: BrokerClient,
    verbose: bool = False,
) -> str | None:
    dest = Path(dest)
    body = broker_models.Listing(path=src, recurse=recurse, include_stats=False)
    try:
        files = broker_client.files.list_dir(body)
    except BrokerError as e:
        raise CLIError(Messages.E018, e)

    src_exists = files.exists
    src_is_file = files.is_file
    src_is_dir = src_exists and not src_is_file
    src_has_trailing_slash = src.endswith("/")

    dest_is_dir = dest.is_dir()
    dest_is_file = dest.is_file()

    #     print(
    #         f"""
    # _download({src=},{dest=}):
    #            {files.paths=}
    #            {src_is_file=}
    #            {src_is_dir=}
    #            {src_has_trailing_slash=}
    #
    #            {dest_exists=}
    #            {dest_is_dir=}
    #            {dest_is_file=}
    # """.strip()
    #     file=sys.stderr,
    #     )

    if not src_exists and not files.paths:
        raise CLIError("No such file or directory", src)

    if dest_is_file:
        if src_is_file and not overwrite:
            raise CLIError(Messages.E020, dest)
        elif src_is_dir:
            raise CLIError("A file already exists at the destination path")

    if src_is_dir and not recurse:
        raise CLIError(Messages.E016.format(verb="download"), src)

    if src_is_file:
        assert len(files.paths) == 1
        src_file = files.paths[0]
        if dest_is_file:
            dest_file = dest
        elif dest_is_dir:
            dest_file = dest / posixpath.basename(src_file)
        else:
            dest_file = dest
        dest_base = dest_file.parent
        create_base = False
        paths = [(src_file, dest_file)]
    else:
        if dest_is_dir and src_has_trailing_slash:
            # `cp src/ dest_exists/` copies `src/**` into `dest_exists/**`
            dest_base = dest
            create_base = False
        elif dest_is_dir and not src_has_trailing_slash:
            # `cp src dest_exists/` copies `src/**` into `dest_exists/src/**`
            dest_base = dest / posixpath.basename(src)
            create_base = True
        else:
            # `cp src dest_noexists/` copies `src/**` into `dest_noexists/**`
            create_base = True
            dest_base = dest
        src_base = _infer_remote_source_base(src, files.paths)

        paths = [
            (src_file, dest_base / posixpath.relpath(src_file, src_base))
            for src_file in files.paths
        ]

    if not dest_base.is_dir():
        if make_dirs:
            dest_base.parent.mkdir(exist_ok=True, parents=True)
        elif create_base:
            dest_base.mkdir(exist_ok=True)
        else:
            raise CLIError(Messages.E021, dest_base)
    #     print(
    #         f"""
    #            {paths=}
    # """.strip()
    #     )

    with Progress(paths) as paths:
        for src_file, dest_file in paths:
            # Determine local destination for this file
            if dest_file.exists() and not overwrite:
                # We should only reach this check for recursive copies, so
                # rather than fail halfway through we just warn
                msg.warn(f"{dest_file} exists, skipping...")
                continue
            body = broker_models.Downloading(target=src_file)
            try:
                content = broker_client.files.download(body)
            except BrokerError as e:
                raise CLIError(Messages.E017.format(verb="download"), e)

            # Save content to local file
            dest_file.parent.mkdir(exist_ok=True, parents=True)
            if verbose:
                msg.info(f"writing {src_file} -> {dest_file}")
            with open(dest_file, "wb") as f:
                f.write(content.read())


def _copy_on_remote(
    *,
    src: str,
    dest: str,
    make_dirs: bool,
    overwrite: bool,
    recurse: bool,
    broker_client: BrokerClient,
) -> str | None:
    if recurse:
        body = broker_models.Copying(
            src=src, dest=dest, make_dirs=make_dirs, overwrite=overwrite
        )
        try:
            plan = broker_client.files.plan_directory_copy(body)
        except BrokerError as e:
            raise CLIError(Messages.E022, e)
        copies = plan.copy_
    else:
        copies = [
            broker_models.Copying(
                src=src, dest=dest, make_dirs=make_dirs, overwrite=overwrite
            )
        ]
    with Progress(copies) as copies:
        for file_copy_plan in copies:
            body = broker_models.Copying(
                src=file_copy_plan.src,
                dest=file_copy_plan.dest,
                make_dirs=make_dirs,
                overwrite=overwrite,
            )
            try:
                broker_client.files.copy(body)
            except BrokerError as e:
                raise CLIError(Messages.E017.format(verb="copy"), e)



"""ptc infra setup — install cluster infrastructure prerequisites."""

from typing import Optional

from radicli import Arg
from wasabi import msg

from ...cli import cli
from ._helpers import ensure_cert_manager, ensure_traefik


@cli.subcommand(
    "infra",
    "setup",
    traefik=Arg("--traefik", help="Install Traefik ingress controller"),
    cert_manager=Arg("--cert-manager", help="Install cert-manager with Let's Encrypt"),
    acme_email=Arg(
        "--acme-email", help="Email for ACME/Let's Encrypt certificate registration"
    ),
    public_ip=Arg("--public-ip", help="Static IP for the Traefik load balancer"),
)
def setup(
    traefik: bool = False,
    cert_manager: bool = False,
    acme_email: Optional[str] = None,
    public_ip: Optional[str] = None,
) -> None:
    """Install cluster infrastructure prerequisites (Traefik, cert-manager).

    These operations are idempotent and safe to re-run.
    """
    if traefik:
        ensure_traefik(public_ip=public_ip or "")
        msg.good("Traefik installed")
    if cert_manager or acme_email:
        ensure_cert_manager()
        msg.good("cert-manager installed")



import builtins
import json
from collections.abc import Sequence
from typing import Optional, Union
from uuid import UUID

from radicli.util import Arg
from wasabi import msg

from prodigy_teams_pam_sdk.models import AssetCreating, AssetDetail, AssetSummary

from ..cli import cli
from ..errors import CLIError, ProdigyTeamsErrors
from ..messages import Messages
from ..query import resolve_asset, resolve_asset_id
from ..ui import print_info_table, print_mutation_result, print_table_with_select
from ..util import resolve_remote_path
from ._state import get_auth_state


@cli.subcommand(
    "assets",
    "list",
    # fmt: off
    select=Arg(
        "--select",
        help=Messages.select.format(opts=builtins.list(AssetSummary.model_fields)),
    ),
    as_json=Arg("--json", help=Messages.as_json),
    # fmt: on
)
def list(
    select: builtins.list[str] = ["id", "name", "kind"], as_json: bool = False
) -> Sequence[AssetSummary]:
    """List all assets on the cluster registered with Ellf"""
    client = get_auth_state().client
    items = builtins.list(client.asset.all())
    print_table_with_select(items, select=select, as_json=as_json)
    return items


@cli.subcommand(
    "assets",
    "info",
    name_or_id=Arg(help=Messages.name_or_id.format(noun="asset")),
    project_id=Arg(help=Messages.project_id.format(noun="asset")),
    cluster_id=Arg(help=Messages.cluster_id.format(noun="asset")),
    select=Arg(
        "--select",
        help=Messages.select.format(opts=builtins.list(AssetDetail.model_fields)),
    ),
    as_json=Arg("--json", help=Messages.as_json),
)
def info(
    name_or_id: Union[str, UUID],
    project_id: Optional[UUID] = None,
    cluster_id: Optional[UUID] = None,
    select: Optional[builtins.list[str]] = None,
    as_json: bool = False,
) -> AssetDetail:
    """
    Get detailed info for an asset uploaded to the cluster and registered
    with Ellf
    """
    res = resolve_asset(name_or_id, broker_id=cluster_id, project_id=project_id)
    print_info_table(res, as_json=as_json, select=select)
    return res


@cli.subcommand(
    "assets",
    "create",
    name=Arg(help=Messages.name.format(noun="asset")),
    kind=Arg("--kind", help=Messages.asset_kind),
    path=Arg(help=Messages.path.format(noun="asset")),
    version=Arg("--version", help=Messages.version.format(noun="asset")),
    meta=Arg("--meta", help=Messages.asset_meta),
    exists_ok=Arg("--exists-ok", help=Messages.exists_ok),
    as_json=Arg("--json", help=Messages.as_json),
)
def create(
    kind: str,
    name: str,
    path: str,
    version: str = "0.0.0",
    meta: str = "{}",
    exists_ok: bool = False,
    as_json: bool = False,
) -> UUID | None:
    """
    Create an asset on the cluster and register it with Ellf. Assets
    point to files or directories you control. The Ellf server only has
    a reference to them. This command doesn't transfer any data. See `ptc files`
    for utilities to transfer files to and from your cluster
    """
    try:
        _ = json.loads(meta)
    except json.JSONDecodeError:
        raise CLIError(Messages.E121, str(meta))
    auth = get_auth_state()
    client = auth.client
    resolved_path = resolve_remote_path(client, path, auth.broker_host)
    try:
        asset = client.asset.create(
            AssetCreating(
                broker_id=auth.broker_id,
                name=name,
                kind=kind,
                version=version,
                meta=json.loads(meta),
                path=resolved_path,
            )
        )
    except ProdigyTeamsErrors.AssetExists:
        if exists_ok:
            msg.info(Messages.T001.format(noun="asset", name=name))
            return None
        raise CLIError(Messages.E002.format(noun="asset", name=name))
    assert asset.path is not None
    print_mutation_result(
        {"id": str(asset.id), "name": asset.name},
        Messages.T002.format(noun="asset", name=asset.name),
        as_json=as_json,
    )
    return asset.id


@cli.subcommand(
    "assets",
    "delete",
    name_or_id=Arg(help=Messages.name_or_id.format(noun="asset")),
    project_id=Arg(help=Messages.project_id.format(noun="asset")),
    cluster_id=Arg(help=Messages.cluster_id.format(noun="asset")),
    as_json=Arg("--json", help=Messages.as_json),
)
def delete(
    name_or_id: Union[str, UUID],
    cluster_id: Optional[UUID] = None,
    project_id: Optional[UUID] = None,
    as_json: bool = False,
) -> UUID:
    """Delete an asset registered with Ellf"""
    asset_id = resolve_asset_id(name_or_id, broker_id=cluster_id, project_id=project_id)
    auth = get_auth_state()
    try:
        auth.client.asset.delete(id=asset_id)
    except ProdigyTeamsErrors.AssetNotFound:
        raise CLIError(Messages.E006.format(noun="asset", name=name_or_id))
    except ProdigyTeamsErrors.AssetForbiddenDelete:
        raise CLIError(Messages.E007.format(noun="asset", name=name_or_id))
    else:
        print_mutation_result(
            {"id": str(asset_id), "deleted": True},
            Messages.T003.format(noun="asset", name=name_or_id),
            as_json=as_json,
        )
    return asset_id
